DEFAULT_FMT = '{asctime:<23s} [{levelname:<8s}] {message} ({name})'
SHORT_FMT = '{asctime:<23s} [{levelname:<8s}] {message}'
PID_FMT = '{asctime:<23s} [{levelname:<8s}] {message} ({name} | PID-{process})'
RAW_FMT = '{message}'
